/* matthew's sprite engine */
/* compile with -lvgagl -lvga */
/* overlays sprites onto a background and displays,
   similar to Commodore 64 VIC chip */

/* $Id: sprite.c,v 1.6 1996/07/13 21:48:35 tom Exp $ */

#include <stdio.h>
#include <string.h>
#include <vga.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>

#include "sprite.h"

signed int spritex[1000];
signed int spritey[1000];
int spritee[1000];
int spritet[1000];
int spritez[1000];
/* arrays containing information about each sprite - x and y co-ordinates
and its type number. type -1 means no object. 1000 sprites allowed at once */

signed int dummy;
int dummy2; /* experimental variables */
int dummy3;

extern int deadcount, rapidfirecount;

/* sprite data arrays */
char sprite_image[1000][16000]; /* 1000 sprite types, sprites can be up to 160x100 pixels */

char *workvirtual; /* working virtual screen */
char *backvirtual; /* background virtual screen */

void spr_init (void)
{
	int i, j;
	FILE *palfile;
	unsigned int r, g, b;

	vga_init (); /* initialise svgalib */
	vga_setmode (G320x200x256); /* switch to 320x200 VGA mode */
	workvirtual = malloc (64000);
	backvirtual = malloc (64000);
	
	/* SPRITE IMAGE DATA */ 
	/* the first 2 bytes are the x and y sizes of the sprite repsectively.
	   Make sure you get these right! After that each byte represents a colour
	   from the VGA palette. Colour 255 (hex ff) means 'transparent'. */

	/* fill spritedata array with 'blanks' */
	for (dummy2 = 0; dummy2 < 1000; dummy2++)
		{
			spritex[dummy2] = 255;
			spritey[dummy2] = 255;
			spritet[dummy2] = 255;
			spritee[dummy2] = 0;
		}
	
	/* install palette */
	if ((palfile = fopen ("palette.pal", "r")) != NULL)
			for (i = 0; i < 256; i++)
				{
					fscanf (palfile, "%u %u %u\n", &r, &g, &b);
					vga_setpalette (i, r, g, b);
				}
	else
		printf ("WARNING: Could not load palette.\n");
}

void spr_loadspr (char *fname, int num)
{
	FILE *tmp;
	
	if ((tmp = fopen (fname, "r")) == NULL)
		return;
	fread (&sprite_image[num], 2, 1, tmp);	
	fread (&sprite_image[num][2], sprite_image[num][0],
	       sprite_image[num][1], tmp);
}
                                                        
void spr_add (int nx, int ny, int nt) /* add a new sprite */
{
	int msv;

	/* find first unused sprite space */
	for (msv = 0 ; (spritet[msv] != 255) && (msv < 1000); msv++ ) ;

	spritex[msv] = nx;
	spritey[msv] = ny;
	spritet[msv] = nt;
}

int spr_add2 (int nx, int ny, int nt) /* add a new sprite, return number */
{
	int msv;

	/* find first unused sprite space */
	for (msv = 0 ; (spritet[msv] != 255) && (msv < 1000); msv++ ) ;

	spritex[msv] = nx;
	spritey[msv] = ny;
	spritet[msv] = nt;
	return(msv);
}

void spr_kill (int kn) /* delete a sprite from the screen */
{
	spritet[kn] = 255;
}

void spr_move (int mn,int mx, int my) /* move a sprite to a new x,y position */
{
	spritex[mn] = mx ; spritey[mn] = my ;
}

void spr_type (int mn, int mt) /* change a sprite's type */
{
	spritet[mn] = mt;
}

void setpixel (int x, int y, int c, char *base)
{
	if (x>=0 && x<320 && y>=0 && y<200) *(base + y * 320 + x) = c;
}

void spr_update (void) /* update screen */
{
	int x, y, c, crap1, crap2, crap3, crap4, crap5, crap6, wx, wy;
	char *src, *dest;
	
	memcpy (workvirtual, backvirtual, 64000);
	
	for (crap6 = 0; crap6 < 1000; crap6++)
		{
			crap1 = spritet[crap6];

			/* is there a sprite to be drawn? */
			if (crap1 != 255 && crap1 != 6 && (crap1 != 0 && deadcount != 0))
				{
					for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
						/* y loop */
						{
							for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
								/* x loop */
								{
									x = crap3;
									y = crap2;
									wx = sprite_image[crap1][0];
									wy = sprite_image[crap1][1];
									c = sprite_image[crap1][wx*crap2+crap3+2];
									if (c != -1)
										setpixel(x + spritex[crap6], y + spritey[crap6], c, workvirtual);
								}
						}
				}
			else if (crap1 == 6)
				{
					for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
						/* y loop */
						{
							dest = workvirtual+((spritey[crap6] + crap2) * 320 + spritex[crap6]);
							src = &(sprite_image[crap1][crap2 * 6 + 2]);
							memcpy (dest, src, 6);
						}
				}
		}
	/* vga_waitretrace(); */
	memcpy (graph_mem, workvirtual, 64000);
}

void spr_update_reverse (void) /* update screen, reverse sprite order */
{
	int x, y, c, crap1, crap2, crap3, crap4, crap5, crap6, wx, wy;
	char *src, *dest;

	memcpy (workvirtual, backvirtual, 64000);
	for (crap6 = 999; crap6 >= 0; crap6--)
	{
		crap1=spritet[crap6];

		/* is there a sprite to be drawn? */
		if (crap1 != 255 && crap1 != 6 && crap1 != 0 && crap1 != 2 && crap1 != 11 && crap1 != 13)
			{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], c, workvirtual);
							}
					}
			}
		else if (crap1 == 6)
			{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						dest = workvirtual + ((spritey[crap6] + crap2) * 320 + spritex[crap6]);
						src = &(sprite_image[crap1][crap2 * 6 + 2]);
						memcpy (dest, src, 6);
					}
			}
		else if (crap1 == 0)
			{
			if (deadcount==0 && spritee[0]==0)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], c, workvirtual);
							}
					}
				}
			else if (deadcount==0 && spritee[0]!=0)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (rnd(spritee[0])!=0 && c!=-1) c=-1;
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], c, workvirtual);
							}
					}
				}
			else if (deadcount>50)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
/*								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], rnd(256), workvirtual);
*/								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									if (rnd(deadcount/50)!=0) setpixel (x + spritex[crap6],y + spritey[crap6], rnd(256), workvirtual);
							}
					}
				}
			}
		else if (crap1==2)
			{
				if (spritee[crap6]<3)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], c, workvirtual);
							}
					}
				}
				else if (spritee[crap6]>2)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3 + spritex[crap6];
								y = crap2 + spritey[crap6];
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel ((x+srnd(spritee[crap6])/2.5), (y+srnd(spritee[crap6])/2.5), c, workvirtual);
							}
					}
				}
			}
		else if (crap1==11)
			{
				if (spritee[crap6]<3)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], c, workvirtual);
							}
					}
				}
				else if (spritee[crap6]>2)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3 + spritex[crap6];
								y = crap2 + spritey[crap6];
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x, (y+srnd(spritee[crap6])/1.5), c, workvirtual);
							}
					}
				}
			}			
		else if (crap1==13)
			{
				if (spritee[crap6]==0)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3;
								y = crap2;
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								if (c != -1)
									setpixel (x + spritex[crap6],y + spritey[crap6], c, workvirtual);
							}
					}
				}
				else if (spritee[crap6]!=0)
				{
				for (crap2 = 0; crap2 < (sprite_image[crap1][1]); crap2++)
					/* y loop */
					{
						for (crap3 = 0; crap3 < (sprite_image[crap1][0]); crap3++)
							/* x loop */
							{
								x = crap3 + spritex[crap6];
								y = crap2 + spritey[crap6];
								wx = sprite_image[crap1][0];
								wy = sprite_image[crap1][1];
								c = sprite_image[crap1][wx * crap2 + crap3 + 2];
								//if (c != -1 && rnd(spritee[crap6])!=0) c=0;
								/*if (spritee[crap6]>20 && c!=-1 && rnd(52-spritee[crap6])==0) c=0;*/
								if (c != -1)
									setpixel (x, (y-rnd(spritee[crap6])), c, workvirtual);
							}
					}
				}
			}			
	}
	if (rapidfirecount!=0)
		{
		crap1 = rapidfirecount/20;
		for (crap2=250;crap2<250+crap1;crap2++)
			{
			setpixel(crap2,0,184,workvirtual);
			setpixel(crap2,1,180,workvirtual);
			setpixel(crap2,2,176,workvirtual);
			setpixel(crap2,3,180,workvirtual);
			setpixel(crap2,4,184,workvirtual);
			}
		}
	if (deadcount!=0)
		{
		crap1 = deadcount/6;
		for (crap2=300;crap2>300-crap1;crap2--)
			{
			setpixel(crap2,0,190,workvirtual);
			setpixel(crap2,1,186,workvirtual);
			setpixel(crap2,2,182,workvirtual);
			setpixel(crap2,3,186,workvirtual);
			setpixel(crap2,4,190,workvirtual);
			}
		}
	vga_waitretrace ();
	memcpy (graph_mem, workvirtual, 64000);
}

int spr_pixelcollide (int x, int y)	/* is pixel x,y occupied by a sprite? if so, which? */
{
	int zn, sn, ax, bx, ay, by, sx, sy;

	zn = 1001; /* 1001 = no sprite found */
	
	/* loop through sprites */
	for (sn = 0; sn < 1000; sn++)
		{
			if (spritet[sn] != 255)
				{
					sx = sprite_image[spritet[sn]][0];
					sy = sprite_image[spritet[sn]][1];
					
					/* y loop */
					for (ay = 0; ay < sy; ay++)
						{
							/* x loop */
							for (ax = 0; ax < sx; ax++) 
								{
									if (((spritex[sn] + ax) == x) &&
									    ((spritey[sn] + ay) == y) &&
									    (zn == 1001) &&
									    (sprite_image[sn][(ax * sx) + ay + 2] != -1))
										zn = sn;
								}
						}
				}
		}
	
	return zn;
}

int spr_pix2collide (int x, int y)	/* like spr_pixelcollide but doesn't take transparent pixels into account. */
{
	int zn, sn, ax, bx, ay, by, sx, sy;

	zn = 1001; /* 1001 = no sprite found */
	
	/* loop through sprites */
	for (sn = 0; sn < 1000; sn++)
		{
			if (spritet[sn] != 255)
				{
					sx = sprite_image[spritet[sn]][0];
					sy = sprite_image[spritet[sn]][1];
					/* y loop */
					for (ay = 0; ay < sy; ay++)
						{
							/* x loop */
							for (ax = 0; ax < sx; ax++)
								{
									if (((spritex[sn] + ax) == x) &&
									    ((spritey[sn] + ay) == y) &&
									    (zn == 1001))
										zn = sn;
								}
						}
				}
		}
	
	return zn;
}

int spr_pix3collide (int x, int y, int excl)	/* like spr_pix2collide but doesn't take transparent pixels into account and excludes the specified sprite. */
{
	int zn, sn, ax, bx, ay, by, sx, sy;

	zn = 1001; /* 1001 = no sprite found */
	
	/* loop through sprites */
	for (sn = 0; sn < 1000; sn++)
		{
			if (spritet[sn] != 255)
				{
					sx = sprite_image[spritet[sn]][0];
					sy = sprite_image[spritet[sn]][1];
					
					/* y loop */
					for (ay = 0; ay < sy; ay++)
						{
							/* x loop */
							for (ax = 0; ax < sx; ax++)
								{
									if (((spritex[sn] + ax) == x) &&
									    ((spritey[sn] + ay) == y) &&
									    (zn == 1001) && (sn != excl))
										zn = sn;
								}
						}
				}
		}
	
	return zn;
}

int spr_spritecollide (int s)	/* does sprite s collide with another sprite? if so, which? WARNING - THIS IS SKODA */
{
	int x, y, bx, by;

	x = sprite_image[spritet[s]][0];
	y = sprite_image[spritet[s]][1];
	bx = spritex[s];
	by = spritey[s];
	x = x / 2; y = y / 2;
	return spr_pix3collide (bx+x, by+y, s);
}

int spr_xsize (int s)
{
	return sprite_image[spritet[s]][0];
}

int spr_ysize (int s)
{
	return sprite_image[spritet[s]][1];
}

void spr_dump (void) /* dump sprite info to standard output */
{
	for (dummy = 0; dummy < 1000; dummy++)
		{
			printf ("%d:x=%d,y=%d,t=%d\n", dummy, spritex[dummy], spritey[dummy], spritet[dummy]);
		}
}
